function [H,conn,M,sd,SigmaS,F,gr] = gsom(X,k,d, speed_up,plo)
% 
% A probabilistic SOM implementaion, using EM optimization
%
% [H,conn,M,sd,SigmaS,F] = gsom(X,k,d,speed_up,plo)
%
% X (N x D) - N D-dimenional data 
% k integer - number of units along each dimension of latent space
% d integer - number of latent dimensions
% speed_up  - number of nearest nodes to check as winner, check all if zero
% plo       - do plotting if non-zero
%
% H         neighborhood function
% conn      matrix with ones for neighbors in the lattice
% M         means of units
% sd        squared distances between units and data points
% SigmaS    sigma squared: the variance
% F         trace of free-energy
%
% by Jakob Verbeek 2002, see http://www.science.uva.nl/~jverbeek
% Only for non-commercial use !
%


% DO INITIALIZATION WORK 

[N,D] = size(X);
f     = 1;   % --> f is width of neighborhood function (neurons are one unit separated) 

if plo  figure(1);clf;set(1,'Doub','on'); end

% init neighborhood function 
if d==1;        gr = [1:k]';
elseif d==2;    [x,y]=meshgrid(1:k,1:k);  gr = [ x(:) y(:)];  end
H = exp( -sqdist(gr',gr') / (2*(f*k)^2)); k=size(gr,1);
H = H ./ repmat(sum(H,1),k,1);

conn = sqdist(gr',gr')<= 1;
%[tmp,tmpconn] = sort(H,1);tmpconn = tmpconn(end-d*2:end-1,:);
%for i=1:k; conn(i,tmpconn(:,i))=1;end;conn=conn+conn';conn=conn>0;size(conn)

entr = sum(-H.*log(H+realmin),1);

% Init neurons

W  = floor(rand(N,1)*k)+1; %initialize winner vector at random
q  = H(:,W);
M  = diag(sum(q,2).^-1)*q*X; 
sd = sqdist(X',M');
SigmaS = mean(min(sd,[],2));

% Init some counters

iter=0;
old_obj  = -sum(sd.*q',2)/(2*SigmaS)+ entr(W)' - D*log(SigmaS)/2;
F=[0];first_bunch=1;
rate   = 1.1;    % --> shrink nh-function, make factor larger for fast shrink

while H(1,1)<.8      % ==== loop over EM steps ======   stop loops if neighborhood function puts more than .95 mass on center 

  iter = iter+1;
 
  [W,obj] = e_step(H,speed_up,W,old_obj,SigmaS,entr,sd,D); % ==== E-step ===
  
  if abs(sum(obj)/sum(old_obj)-1)<1e-5; [H, F, entr,obj] = cool(H,rate,k,W,sd,SigmaS,D,F,0*plo);end    % ====== Cool after convergence ====
  q = H(:,W);fprintf('.');  % === set q as nh-func of winner

  [M,SigmaS,sd] = m_step(X,q);      % ========== M step   =========
  
  if plo                                                     % ========== Plotting =========
      if first_bunch & length(F)>2;F=F(2:end) ;first_bunch=0;end % throw away first two values of F (possibly very high) since they may screw-up the plot
      plot_it(X,M,1*sqrt(SigmaS),conn); 
  end    

  old_obj=obj;

end


